"""This module contains the managers for the ``django-discover`` app"""

import io
import os
from zipfile import ZipFile, ZIP_LZMA

from django.conf import settings
from django.core.serializers import deserialize, serialize
from django.db import transaction
from django.db.models import (
    Manager,
    Q,
    QuerySet
)

import requests
from psycopg2.extras import NumericRange


class DiscoverItemQueryset(QuerySet):
    """QuerySet for the ``DiscoverItem`` model"""

    def for_language(self, language):
        """Filter the discover items for the given language

        Parameters
        ----------
        language: str
            The language to filter on

        Returns
        -------
        DiscoverItemQueryset
            The filtered Queryset

        """

        return self.filter(language=language)

    def for_version(self, major_version, minor_version=0, hotfix_version=0):
        """Filter the discover items for the given version

        The query below can be decomposed like this:
        - Return all the item for which the major versions field is not filled
        - Return the items for which the major version field is filled IF:
          - The given version is in the bounds
          - AND The minor version is in the bounds (if filled)
          - AND The hotfix version is in the bounds (if filled)

        Parameters
        ----------
        major_version: int
            The major version
        minor_version: int
            The minor version
        hotfix_version: int
            The hotfix version

        Returns
        -------
        DiscoverItemQueryset
            The filtered Queryset

        """

        return self.filter(
            # The major version was not filled
            Q(major_versions__isnull=True) |
            Q(
                # The major versions contains the current version
                Q(
                    major_versions__contains=NumericRange(
                        major_version,
                        major_version,
                        bounds='[]'
                    )
                ) &
                Q(
                    Q(minor_versions__isnull=True) |
                    Q(
                        minor_versions__isnull=False,
                        minor_versions__contains=NumericRange(
                            minor_version,
                            minor_version,
                            bounds='[]'
                        )
                    )
                ) &
                Q(
                    Q(hotfix_versions__isnull=True) |
                    Q(
                        hotfix_versions__isnull=False,
                        hotfix_versions__contains=NumericRange(
                            hotfix_version,
                            hotfix_version,
                            bounds='[]'
                        )
                    )
                )
            )
        )


class DiscoverItemManager(Manager.from_queryset(DiscoverItemQueryset)):
    """Manager for the ``DiscoverItem`` model"""

    @staticmethod
    def get_zipped_discover_items(discover_items_queryset):
        """Get a ZipFile containing the serialized discover items queryset

        Parameters
        ----------
        discover_items_queryset: DiscoverItemQueryset
            The filtered queryset containing the discover items to serialize

        Returns
        -------
        io.BytesIO
            A buffer of a ZipFile containing the following elements:
            - A json file containing the serialized discover items
            - The media of the serialized items

        """

        # Build the ZipFile
        zip_buffer = io.BytesIO()
        zip_file = ZipFile(zip_buffer, 'w', ZIP_LZMA)

        # Copy the serialized discover items
        serialized_items = serialize('json', discover_items_queryset)
        zip_file.writestr('discover_items.json', serialized_items)

        # Copy the media
        medias_list = discover_items_queryset.values_list('image', flat=True)
        for media in medias_list:
            file_path = os.path.join(settings.MEDIA_ROOT, media)
            zip_file.write(file_path, media)

        zip_file.close()

        return zip_buffer

    def load_zipped_discover_items(self):
        """Load an archive containing serialized discover items

        Raises
        ------
        Exception
            If something went wrong

        """

        discover_items_media_path = os.path.join(
            settings.MEDIA_ROOT,
            'django_discover'
        )

        try:
            with transaction.atomic():
                # 1 - Get the discover items archive from the Command center
                response = requests.get(settings.COMMAND_CENTER_URL, stream=True)

                try:
                    response.raise_for_status()
                except requests.HTTPError:
                    # A 404 error means no discover items were available
                    if response.status_code == 404:
                        return False
                    else:
                        # If we encounter another error, re-raise the ``HTTPError``
                        raise

                zip_file = ZipFile(io.BytesIO(response.content), 'r', ZIP_LZMA)

                # 2 - Erase all of the actual discover items
                self.get_queryset().all().delete()

                # 3 - Rename the path containing the media of the current discover items
                try:
                    os.rename(discover_items_media_path, discover_items_media_path + '_save')
                except OSError:
                    pass

                # 4 - Extract the media files of the new discover items
                for file_name in zip_file.namelist():
                    if file_name == 'discover_items.json':
                        continue

                    zip_file.extract(file_name, settings.MEDIA_ROOT)

                # 5 - Load the new discover items objects from the json file
                for discover_item in deserialize('json', zip_file.read('discover_items.json')):
                    discover_item.save()

        except Exception as exception:  # pylint: disable=broad-except

            # Some things went wrong, move back the media...
            try:
                os.rename(discover_items_media_path + '_save', discover_items_media_path)
            except OSError:
                pass

            # ...And raise
            raise ValueError(
                'Error during the discover items fetching. The exception was "%s"' % exception
            )

        return True






"""This module contains the views of the ``django_discover`` app"""

from django.http import Http404, HttpResponse
from django.views.generic import ListView

from .constants import DISCOVER_ITEMS_ARCHIVE_NAME
from .models import DiscoverItem


class DiscoverItemsArchiveView(ListView):
    """Django List View returning an archive for the provided version"""

    model = DiscoverItem

    def get_queryset(self):
        """Filter the ``DiscoverItem``s accordingly to the url param "version".

        Raises
        ------
        Http404
            If the queryset is empty for the given version

        """

        queryset = super().get_queryset().for_version(
            *[
                int(version_part)
                for version_part in self.kwargs['version'].split('.')
            ]
        )

        if not queryset.exists():
            raise Http404

        return queryset

    def render_to_response(self, context, **response_kwargs):
        """Return a ZipFile containing the discover items"""

        zip_buffer = DiscoverItem.objects.get_zipped_discover_items(
            self.object_list
        )

        response = HttpResponse(content_type='application/x-zip-compressed')
        response['Content-Disposition'] = 'attachment; filename=%s' % DISCOVER_ITEMS_ARCHIVE_NAME

        response.write(zip_buffer.getvalue())

        return response



"""This module contains the fields for the ``django-discover`` app"""

from django.conf import settings
from django.db import models


class LanguageField(models.CharField):
    """
    A language field for Django models.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('max_length', 5)
        kwargs.setdefault('choices', sorted(settings.LANGUAGES, key=lambda x: x[1]))
        super().__init__(*args, **kwargs)



"""
``django-discover`` provides ``DiscoverItem`` model, that allow displaying an introduction
to an existent or a new feature to the user.
"""

__author__ = 'Clément "clemfromspace" Denoix'
__contact__ = 'clement.denoix@magency.fr'
__homepage__ = 'https://github.com/magencydigital/'
__license__ = 'BSD'
__version__ = '0.0.1a3'
VERSION = tuple(__version__.split('.'))




# -*- coding: utf-8 -*-
# Generated by Django 1.11.3 on 2018-01-19 02:37
from __future__ import unicode_literals

import django.contrib.postgres.fields.ranges
from django.db import migrations, models
import django_discover.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='DiscoverItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('language', django_discover.fields.LanguageField(choices=[('bg', 'Bulgarian'), ('cs', 'Czech'), ('da', 'Dansk'), ('de', 'Deutsch'), ('de-AT', 'Deutsch (Austria)'), ('de-CH', 'Deutsch (Swiss)'), ('nl-BE', 'Dutch (Belgium)'), ('en', 'English'), ('en-GB', 'English (Britain)'), ('es', 'Español'), ('fr', 'Français'), ('fr-BE', 'Français (Belgium)'), ('fr-CH', 'Français (Swiss)'), ('it', 'Italiano'), ('it-CH', 'Italiano (Swiss)'), ('hu', 'Magyar'), ('nl', 'Nederlands'), ('no', 'Norsk'), ('pl', 'Polski'), ('pt', 'Portuguese'), ('sk', 'Slovak'), ('fi', 'Suomi'), ('sv', 'Svenska'), ('el', 'Ελληνικά'), ('ru', 'Русский'), ('ja', '日本語')], help_text='The language of this discover item', max_length=5, verbose_name='Language')),
                ('title', models.CharField(max_length=30, verbose_name='Title')),
                ('slug', models.SlugField(max_length=500, verbose_name='Slug')),
                ('image', models.ImageField(upload_to='django_discover', verbose_name='Image')),
                ('short_description', models.CharField(max_length=50, verbose_name='Short description')),
                ('long_description', models.TextField(verbose_name='Long description')),
                ('link_title', models.CharField(blank=True, max_length=20, null=True, verbose_name='Link title')),
                ('link_url', models.CharField(blank=True, max_length=500, null=True, verbose_name='Link url')),
                ('tag', models.CharField(blank=True, choices=[('new', 'New'), ('important', 'Important')], max_length=20, null=True, verbose_name='Tag')),
                ('category', models.CharField(blank=True, choices=[('resources', 'Resources'), ('sessions', 'Sessions'), ('users', 'Users')], max_length=20, null=True, verbose_name='Category')),
                ('major_versions', django.contrib.postgres.fields.ranges.IntegerRangeField(blank=True, null=True, verbose_name='Major version range')),
                ('minor_versions', django.contrib.postgres.fields.ranges.IntegerRangeField(blank=True, null=True, verbose_name='Minor version range')),
                ('hotfix_versions', django.contrib.postgres.fields.ranges.IntegerRangeField(blank=True, null=True, verbose_name='Hotfix version range')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Date and time of creation')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Date and time of the last update')),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='discoveritem',
            unique_together=set([('language', 'slug')]),
        ),
    ]
